from collections import defaultdict
import json
import os

import pandas as pd

from tuneconfig.experiment import Experiment
from tuneconfig.trial import Trial


class ExperimentAnalysis:
    """ExperimentAnalysis

    """

    def __init__(self, logdir):
        self.logdir = logdir

        self._trials = {}
        self._params = defaultdict(set)

    @property
    def params(self):
        return {key: sorted(values) for key, values in self._params.items()}

    @property
    def results(self):
        return sorted(self[0][0])

    @property
    def metrics(self):
        return self[0].metrics

    @property
    def size(self):
        n_trials = len(self)
        total_runs = sum(len(trial) for trial in self._trials.values())
        n_runs_per_trial = total_runs / n_trials
        return (n_trials, n_runs_per_trial)

    def info(self):
        n_trials, runs_per_trial = self.size
        print(f"<{self}>")
        print(f"TrialIndex: {n_trials} trials, {runs_per_trial} runs per trial.")
        print(f"ParamIndex: {len(self.params)} parameters.")
        for param, values in self.params.items():
            print(f"  - {param} = [{', '.join(list(map(str, values)))}]")
        print(f"ResultIndex: {len(self.results)} result files.")
        for result, metrics in self.metrics.items():
            print(f"  - {result}({', '.join(metrics)})")

    def setup(self):
        for dirname, subdirs, filenames in os.walk(self.logdir):
            if Experiment.is_trial_dir(dirname):
                trial = Trial.from_directory(dirname)
                self._trials[dirname] = trial

                for key, value in trial.config.items():
                    self._params[key].add(value)

    def get(self, params_values):
        trials = {}
        for name, trial in self._trials.items():
            if all(pv in name for pv in params_values):
                trials[name] = trial
        return trials

    def __str__(self):
        return f"ExperimentAnalysis(logdir='{self.logdir}')"

    def __len__(self):
        return len(self._trials)

    def __getitem__(self, i):
        return list(self._trials.items())[i][1]



from collections import defaultdict

import matplotlib.pyplot as plt


class ExperimentPlotter:
    def __init__(self, analysis):
        self.analysis = analysis

    def plot(
        self,
        targets,
        anchors=None,
        x_axis=None,
        y_axis=None,
        filename=None,
        show_fig=False,
        **kwargs,
    ):
        # filter trials by anchors
        trials = self.analysis.get(anchors)

        # separate trials in grid
        trial_grid = defaultdict(lambda: defaultdict(list))
        xs, ys = set(), set()
        for name, trial in trials.items():
            # get avg-std stats for each required metric
            trial_stats = trial.stats()

            # find required result and metrics
            metrics = []
            for target in targets:
                result, metric = target.split(":")
                metrics.append((target, trial_stats[result][metric]))

            x = trial.config.get(x_axis)
            y = trial.config.get(y_axis)
            trial_grid[x][y].append((trial, metrics))
            if x:
                xs.add(x)
            if y:
                ys.add(y)

        # define number of cells for plots
        nrows, ncols = len(ys), len(xs)
        if nrows == ncols == 0:
            ncols = len(targets)
            nrows = 0
        else:
            ncols = ncols if ncols > 0 else len(targets)
            nrows = nrows if nrows > 0 else len(targets)

        # create figure and subplots
        figsize = kwargs.get("figsize", (7, 5))
        fig, axes = plt.subplots(
            nrows, ncols, sharex=True, sharey=True, figsize=figsize
        )

        # iterate over axes and plot each metric
        for j, x in enumerate(trial_grid):
            for i, y in enumerate(trial_grid[x]):
                trials = trial_grid[x][y]
                metrics = trials[0][1]

                if x_axis and not y_axis:
                    for k, metric in enumerate(metrics):
                        self._plot(axes[k][j], x, y, x_axis, y_axis, metric, **kwargs)
                elif not x_axis and y_axis:
                    for k, metric in enumerate(metrics):
                        self._plot(axes[i][k], x, y, x_axis, y_axis, metric, **kwargs)
                elif not x_axis and not y_axis:
                    for k, metric in enumerate(metrics):
                        self._plot(axes[k], x, y, x_axis, y_axis, metric, **kwargs)
                else:
                    print(i, j)
                    print(axes[i][j])
                    self._plot(axes[k], x, y, x_axis, y_axis, metric, **kwargs)

        # save figure
        if filename:
            fig.savefig(filename, format="pdf")

        # show figure
        if show_fig:
            plt.show()

    def _plot(self, ax, x, y, x_axis, y_axis, metric, **kwargs):
        label, df = metric
        mean, std = df["mean"], df["std"]
        lower = mean - std
        upper = mean + std
        x = range(len(mean))
        ax.plot(x, mean, label=label)
        ax.fill_between(x, lower, upper, alpha=0.3)

        title = f"{x_axis}={x}" if x_axis else ""
        title += ", " if x_axis else ""
        title += f"{y_axis}={y}" if y_axis else ""
        ax.set_title(title)
        ax.set_xlabel(kwargs.get("target_x_axis_label"))

        ax.grid()
        ax.legend()



import json
import os

import pytest

import tuneconfig


def test_grid_search():
    values = tuneconfig.grid_search([0.001, 0.01, 0.1])
    assert isinstance(values, tuneconfig.ParamsIterator)
    assert list(values) == values._lst


def test_tune_config_iterator(config_factory):
    for i, config in enumerate(config_factory):
        assert isinstance(config, dict)
        assert config == config_factory[i]

        for param, value in config.items():
            if param in config_factory._base_dict:
                assert value == config_factory._base_dict[param]
                assert param not in config_factory._params
            else:
                j = config_factory._params.index(param)
                assert value == config_factory._value_instantiations[i][j]
                assert param not in config_factory._base_dict


def test_tune_config_len(config_factory):
    assert len(config_factory) == 18


def test_tune_config_get_item(config_factory):
    for index, config in enumerate(config_factory):
        assert config_factory[index] == config


def test_trial_id(config_factory):
    for config in config_factory:
        name = config_factory._trial_id(config)
        assignments = name.split("/")

        ignore_params = [p for p in config if config_factory._format_fn(p) is None]
        assert len(assignments) == len(config) - len(ignore_params)

        for assignment in assignments:
            param, value = assignment.split("=")

            for p, val in config.items():
                fmt_param = config_factory._format_fn(p)
                if fmt_param == param:
                    assert value == str(val)


def test_tune_config_dump(config_factory):
    tmp = "/tmp/tuneconfig"

    filepaths = config_factory.dump(tmp)
    assert len(filepaths) == len(config_factory)

    for idx, config in enumerate(config_factory):

        filepath = os.path.join(tmp, config_factory._trial_id(config), "config.json")
        assert filepaths[idx] == filepath
        assert os.path.exists(filepath)
        assert os.path.isfile(filepath)

        with open(filepath, "r") as file:
            assert config == json.load(file)


def test_is_config_valid(config_factory):
    ignore = {
        "batch_size": 128,
        "learning_rate": 0.1,
    }

    for config in config_factory:
        for param, value in ignore.items():
            if param not in config or config[param] != value:
                assert config_factory._is_config_valid(config, ignore)

    ignore = [
        {
            "batch_size": 128,
            "learning_rate": 0.1,
        },
        {
            "batch_size": 32,
            "optimizer": "GradientDescent",
            "learning_rate": 0.01,
        },
        {
            "horizon": 100,
            "batch_size": 64
        }
    ]

    for config in config_factory:
        for no_config in ignore:
            for param, value in no_config.items():
                if param not in config or config[param] != value:
                    assert config_factory._is_config_valid(config, no_config)


def test_tune_config_filter(config_factory):
    ignore = [
        {
            "batch_size": 128,
            "learning_rate": 0.1,
        },
        {
            "batch_size": 32,
            "optimizer": "GradientDescent",
            "learning_rate": 0.01,
        },
        {
            "horizon": 100,
            "batch_size": 64
        }
    ]

    tmp = "/tmp/tuneconfig"

    filepaths = config_factory.dump(tmp, ignore=ignore)

    num_valid_configs = 0
    for config in config_factory:

        if not config_factory._is_config_valid(config, ignore):
            filepath = os.path.join(tmp, config_factory._trial_id(config), "config.json")
            assert filepath not in filepaths
        else:
            num_valid_configs += 1

    assert len(filepaths) == num_valid_configs
